import os
import random
import sys
import time
import warnings

import numpy as np
import torch
from torch.backends import cudnn
from tqdm import tqdm, trange

from data import create_dataloader
from utils.logger import Logger
import wandb
import torch.nn as nn


def set_seed(seed):
    cudnn.benchmark = False  # if benchmark=True, deterministic will be False
    cudnn.deterministic = True
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

def update_in(model, sparse_loss_dict, sparse_loss_dict_index,size1,thre):
    idx = 0
    height = size1
    width = size1
    contral_rate = 0.001
    model1 = model.modules()
    crol_tem = 1
    channel_prun = []
    channel_prun_dic = {}
    cnt = 0
    for m in model1:
        if isinstance(m, nn.InstanceNorm2d):
            crol_tem = 1
            idx += 1
            if idx in sparse_loss_dict:
                gamma1 = sparse_loss_dict[idx]
                gamma_max = 0.0
                beta_max = 0.0
                gamma_bord = 0.0
                beta_bord = 0.0
                gamma1_sorted, indices = torch.sort(gamma1)
                tensor_size = len(gamma1)
                for i in range(tensor_size):
                    if sparse_loss_dict_index[idx][i]==1:
                        beta_bord = beta_bord + gamma1[i]
                        beta_max = max(beta_max,gamma1[i])
                    else:
                        gamma_bord = gamma_bord + gamma1[i]
                        gamma_max = max(gamma_max,gamma1[i])

                gamma_bord = gamma_bord*contral_rate
                beta_bord = beta_bord*contral_rate/10
                # gamma_max = gamma_max/10
                # beta_max = beta_max/20
                # gamma_max = gamma_max/8
                # beta_max = beta_max/16
                gamma_max = gamma_max/thre
                beta_max = beta_max/thre/2
                channel_prun = []
                for i in range(tensor_size):
                    tem_sum = gamma1_sorted[:i].sum()
                    if tem_sum > beta_bord and tem_sum > gamma_bord:
                        break
                    if sparse_loss_dict_index[idx][indices[i]]==1:
                        if tem_sum <= beta_bord and gamma1[indices[i]] < beta_max:
                            cnt += 1
                            # print(idx,indices[i])
                            channel_prun.append(indices[i])
                            channel_prun_dic.setdefault(idx,[]).append(indices[i])
                    else:
                        if tem_sum <= gamma_bord and gamma1[indices[i]] <= gamma_max:
                            # m.weight.data[indices[i]] = 0.0
                            # m.bias.data[indices[i]] = 0.0
                            cnt += 1
                            # print(idx,indices[i])
                            channel_prun.append(indices[i])
                            channel_prun_dic.setdefault(idx,[]).append(indices[i])
                # if cnt>0:
                #     print ('layer = {}, zero cnt = {}'.format(idx, cnt))
        if crol_tem == 1 and isinstance(m, nn.Conv2d):        
            width /= m.stride[0]
            height /= m.stride[1]
            for item in channel_prun:
                m.weight.data[:, item, :,:] = 0
            crol_tem = 0 
        if crol_tem == 1 and isinstance(m, nn.ConvTranspose2d):        
            width /= m.stride[0]
            height /= m.stride[1]
            # for item in channel_prun:
            #     m.weight.data[item, :, :,:] = 0
            crol_tem = 0    
            # print ('layer = {}, zero cnt = {}'.format(idx, cnt))

class Trainer:
    def __init__(self, task):
        self.task = task
        from options.distill_options import DistillOptions as Options
        from distillers import create_distiller as create_model

        opt = Options().parse()
        opt.tensorboard_dir = opt.log_dir if opt.tensorboard_dir is None else opt.tensorboard_dir
        print(' '.join(sys.argv))
        if opt.phase != 'train':
            warnings.warn('You are not using training set for %s!!!' % task)
        with open(os.path.join(opt.log_dir, 'opt.txt'), 'a') as f:
            f.write(' '.join(sys.argv) + '\n')
        set_seed(opt.seed)                      # 设置随机种子

        dataloader = create_dataloader(opt)  # create a dataset given opt.dataset_mode and other options
        dataset_size = len(dataloader.dataset)  # get the number of images in the dataset.
        print('The number of training images = %d' % dataset_size)

        model = create_model(opt)  # create a model given opt.model and other options
        model.setup(opt)  # regular setup: load and print networks; create schedulers
        logger = Logger(opt)

        self.opt = opt
        self.dataloader = dataloader
        self.model = model
        self.logger = logger

    def evaluate(self, epoch, iter, message):
        start_time = time.time()
        metrics = self.model.evaluate_model(iter)
        self.logger.print_current_metrics(epoch, iter, metrics, time.time() - start_time)
        self.logger.plot(metrics, iter)
        self.logger.print_info(message)
        self.model.save_networks('latest')

    def start(self):
        torch.backends.cudnn.deterministic = False
        torch.backends.cudnn.benchmark = True

        opt = self.opt
        dataloader = self.dataloader
        model = self.model
        logger = self.logger

        if self.opt.project:
            wandb.init(project=self.opt.project, name=self.opt.name)
            config = wandb.config
            for k, v in sorted(vars(opt).items()):
                setattr(config, k, v)

        start_epoch = opt.epoch_base
        end_epoch = opt.epoch_base + opt.nepochs + opt.nepochs_decay - 1
        total_iter = opt.iter_base
        epoch_tqdm = trange(start_epoch, end_epoch + 1, desc='Epoch      ', position=0, leave=False)
        self.logger.set_progress_bar(epoch_tqdm)
        for epoch in epoch_tqdm:
            display_images = []
            epoch_start_time = time.time()  # timer for entire epoch
            for i, data_i in enumerate(tqdm(dataloader, desc='Batch      ', position=1, leave=False)):
                iter_start_time = time.time()
                total_iter += 1
                model.set_input(data_i)
                if epoch == start_epoch and i == 0:
                    model.profile()
                model.optimize_parameters(total_iter)
                # print(model.lambda_big_t)

                if total_iter % opt.print_freq == 0:
                    losses = model.get_current_losses()
                    logger.print_current_errors(epoch, total_iter, losses, time.time() - iter_start_time)
                    logger.plot(losses, total_iter)
                    if self.opt.project:
                        wandb.log(losses)
                
                update_in(model.netG_teacher_A, model.sparse_loss_dict_t, model.sparse_loss_dict_index_t, opt.crop_size, 20)

                if total_iter % opt.save_latest_freq == 0:
                    if self.opt.project:
                        current_visual_result = model.get_current_visuals()
                        for k, v in current_visual_result.items():
                            display_images.append(wandb.Image(v))
                    self.evaluate(epoch, total_iter,
                                  'Saving the latest model (epoch %d, total_steps %d)' % (epoch, total_iter))
                    if model.is_best:
                        model.save_networks('iter%d' % total_iter)

            if self.opt.project:
                wandb.log({'Image': display_images})

            logger.print_info(
                'End of epoch %d / %d \t Time Taken: %.2f sec' % (epoch, end_epoch, time.time() - epoch_start_time))
            if epoch % opt.save_epoch_freq == 1 or epoch == end_epoch:
                self.evaluate(epoch, total_iter,
                              'Saving the model at the end of epoch %d, iters %d' % (epoch, total_iter))
                if self.task == 'distill' and self.opt.distiller in ['cycleganbest']:
                    model.load_best_teacher()
                    model.optimize_student_parameters()
                    update_in(model.netG_student, model.sparse_loss_dict_s, model.sparse_loss_dict_index_s, opt.crop_size, 4)
                    model.load_latest_teacher()
                model.save_networks(epoch)
            model.update_learning_rate(logger)
